"""Module for managing protein entries derived from FASTA files.

This module provides a ProteinDatabase class for managing protein entries derived from
FASTA files. The database can be used to import FASTA files, access the protein entries
by their identifier, and write the protein entries back to a FASTA file.

Classes:
    AbstractDatabaseEntry (Protocol): Interface for representing a protein entry.
    DatabaseEntry: Representation of a protein entry.
    ProteinDatabase: A database for managing protein entries derived from FASTA files.
"""

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterator, Optional, Protocol

from profasta.parser import get_parser, get_writer
import profasta.io


class AbstractDatabaseEntry(Protocol):
    """A protein entry derived from a protein record in a FASTA file."""

    identifier: str
    header: str
    sequence: str
    header_fields: dict[str, str]


@dataclass
class DatabaseEntry:
    """A protein entry derived from a protein record in a FASTA file."""

    identifier: str
    header: str
    sequence: str
    header_fields: dict[str, str]


class ProteinDatabase:
    """A database for storing protein entries derived from FASTA files.

    This class provides functionality for managing a collection of protein entries
    derived from FASTA files. It allows for importing protein entries from FASTA files,
    adding new entries, and exporting entries back to FASTA format.

    The Python indexing operator `[]` provides access to the protein entries by their
    identifier. Iterating over the database yields the identifiers of all protein
    entries.

    Attributes:
        db: Dictionary mapping protein identifiers to protein entries.
        imported_fasta_files: List of FASTA files that have been imported into the
            database.
    """

    db: dict[str, AbstractDatabaseEntry]
    imported_fasta_files: list[str]

    def __init__(self):
        self.db = {}
        self.imported_fasta_files = []

    def add_fasta(
        self,
        path: str,
        header_parser: str,
        fasta_name: Optional[str] = None,
        overwrite: bool = False,
    ):
        """Add protein entries from a FASTA file to the database.

        Args:
            path: The path to the FASTA file.
            header_parser: The name of the parser to use for parsing the FASTA headers.
                The parser must be registered in the global parser registry.
            fasta_name: Optional name for the FASTA file. If not provided, the filename
                will be used instead.
            overwrite: If True, overwrite an existing entry with the same identifier.
                If False and an entry with the same identifier already exists, a
                KeyError will be raised.
        """
        if fasta_name is None:
            fasta_name = Path(path).name
        self.imported_fasta_files.append(fasta_name)

        parser = get_parser(header_parser)
        with open(path, "r") as file:
            for fasta_record in profasta.io.parse_fasta(file):
                parsed_header = parser.parse(fasta_record.header)
                protein_entry = DatabaseEntry(
                    parsed_header.identifier,
                    parsed_header.header,
                    fasta_record.sequence,
                    parsed_header.header_fields,
                )
                self.add_entry(protein_entry, overwrite)

    def add_entry(self, protein_entry: AbstractDatabaseEntry, overwrite: bool = False):
        """Add a protein entry to the database.

        Args:
            protein_entry: The protein entry to add.
            overwrite: If True, overwrite an existing entry with the same identifier.
                If False and an entry with the same identifier already exists, a
                KeyError will be raised.
        """
        if protein_entry.identifier in self.db and not overwrite:
            raise KeyError(
                f"Identifier {protein_entry.identifier} already in database."
            )

        self.db[protein_entry.identifier] = protein_entry

    def write_fasta(
        self,
        path,
        append: bool = False,
        header_writer: Optional[str] = None,
        line_width: int = 60,
    ):
        """Write all protein entries in the database to a FASTA file.

        Args:
            path: The path to write the FASTA file to.
            append: If False, the file is created or overwritten. If True, the entries
                are appended to an existing file. The default value is False.
            header_writer: The name of the writer to use for generating the FASTA header
                strings from the database entries. If None, the original header strings
                are written to the FASTA file.
            line_width: The number of sequence characters per line, the default value is
                60. If -1, the sequence is not split into multiple lines.
        """
        if header_writer is None:
            fasta_records = list(self.db.values())
        else:
            fasta_records = []
            writer = get_writer(header_writer)
            for protein_entry in self.db.values():
                header = writer.write(protein_entry)

                fasta_records.append(
                    DatabaseEntry("", header, protein_entry.sequence, {})
                )
        file_open_mode = "a" if append else "w"
        with open(path, file_open_mode) as file:
            profasta.io.write_fasta(file, fasta_records, line_width)

    def get(self, identifier: str, default: Any = None) -> DatabaseEntry | Any:
        """Get a protein entry by its identifier or return a default value."""
        return self.db.get(identifier, default)

    def __getitem__(self, identifier) -> AbstractDatabaseEntry:
        return self.db[identifier]

    def __contains__(self, identifier) -> bool:
        return identifier in self.db

    def __iter__(self) -> Iterator[str]:
        return iter(self.db)



"""ProFASTA - A Python library for working with protein containing FASTA files."""

from profasta.db import DatabaseEntry, ProteinDatabase
from profasta.decoy import create_decoy_db
import profasta.parser
import profasta.io


__author__ = "David M. Hollenstein"
__license__ = "MIT"
__version__ = "0.0.3"



"""This module manages parsers and writer for the headers of FASTA records.

This module provides classes for parsing the headers of FASTA records into a structured
format and writing the structured format back to a header string. The default FASTA
header parsers and writers are registered in a global registry, which can be accessed
via the `get_parser` and `get_writer` functions and the name of the parser or writer.
New parsers and writers must be registered via the `register_parser` and
`register_writer` functions before they become available in the other modules.

Classes:
    AbstractParsedHeader (Protocol): Interface for representing a parsed FASTA header.
    ParsedHeader: Representation of a parsed FASTA header.
    HeaderParser (Protocol): Interface for a FASTA header parser.
    HeaderWriter (Protocol): Interface for a FASTA header writer.
    DefaultParser: Default FASTA header parser.
    UniprotParser: Parser for Uniprot FASTA headers.
    UniprotLikeParser: Parser for less strict Uniprot like FASTA headers.
    DefaultWriter: Default FASTA header writer.
    UniprotWriter: Parser for Uniprot FASTA writer.
    UniprotLikeWriter: Parser for less strict Uniprot like FASTA writer.

Functions:
    register_parser: Register a custom FASTA header parser by name.
    get_parser: Get a registered FASTA header parser by name.
    register_writer: Register a custom FASTA header writer by name.
    get_writer: Get a registered FASTA header writer by name.

Constants:
    PARSER_REGISTRY: Dictionary mapping parser names to header parser classes. The
        built-in parsers are registered as "default", "uniprot", "uniprot_like" and can
        be retrieved via the `get_parser` function.
    WRITER_REGISTRY: Dictionary mapping writer names to header writer classes. The
        built-in writers are registered as "default", "uniprot", "uniprot_like" and can
        be retrieved via the `get_writer` function.
"""

from dataclasses import dataclass, field
from typing import Protocol
import re


class AbstractParsedHeader(Protocol):
    """Abstract parsed FASTA header.

    Attributes:
        identifier: The unique identifier of the FASTA entry.
        header: The FASTA header, not containing the starting ">" character.
        header_fields: The parsed header fields as a dictionary.
    """

    identifier: str
    header: str
    header_fields: dict[str, str]


@dataclass
class ParsedHeader:
    """Parsed FASTA header.

    Attributes:
        identifier: The unique identifier of the FASTA entry.
        header: The FASTA header, not containing the starting ">" character.
        header_fields: The parsed header fields as a dictionary.
    """

    identifier: str
    header: str
    header_fields: dict[str, str] = field(default_factory=dict)


class HeaderParser(Protocol):
    """Abstract header parser."""

    @classmethod
    def parse(self, header: str) -> AbstractParsedHeader:
        """Parse a FASTA header string into a ParsedHeader object."""
        ...


class HeaderWriter(Protocol):
    """Abstract header writer."""

    @classmethod
    def write(self, parsed_header: AbstractParsedHeader) -> str:
        """Write a FASTA header string from a ParsedHeader object."""
        ...


class DefaultParser:
    """Default FASTA header parser.

    The `parse` method returns a ParsedHeader object with the identifier being the
    first whitespace-separated word of the header. The rest of the header is stored
    in the "description" field of the `header_fields` dictionary, which might be an
    empty string.

    The `write` method returns the original `header` string from the parsed_header.
    """

    @classmethod
    def parse(cls, header: str) -> ParsedHeader:
        """Parse a FASTA header string into a ParsedHeader object."""
        split_header = header.split(maxsplit=1)
        _id = split_header[0]
        fields = {"description": split_header[1]} if len(split_header) > 1 else {}
        return ParsedHeader(_id, header, fields)


class DefaultWriter:
    """Default FASTA header writer.

    The `write` method returns the original `header` string from the parsed_header.
    """

    @classmethod
    def write(cls, parsed_header: AbstractParsedHeader) -> str:
        """Write a FASTA header string from a ParsedHeader object."""
        return parsed_header.header


class DecoyWriter:
    """A FASTA header writer for decoy entries.

    The `write` method returns the original `header` string from the parsed_header
    prefixed with a tag "rev_" to indicate that the entry is a decoy.
    """

    decoy_tag: str = "rev_"

    @classmethod
    def write(cls, parsed_header: AbstractParsedHeader) -> str:
        """Write a FASTA header string from a ParsedHeader object."""
        return f"{cls.decoy_tag}{parsed_header.header}"


class UniprotParser:
    """Uniprot FASTA header parser."""

    header_pattern = re.compile(
        r"^(?P<db>\w+)\|(?P<id>[-\w]+)\|(?P<entry>\w+)\s+(?P<name>.*?)"
        r"(?:(\s+OS=(?P<OS>[^=]+))|"
        r"(\s+OX=(?P<OX>\d+))|"
        r"(\s+GN=(?P<GN>\S+))|"
        r"(\s+PE=(?P<PE>\d))|"
        r"(\s+SV=(?P<SV>\d+)))*\s*$"
    )

    field_names = {
        "db": "db",
        "id": "identifier",
        "entry": "entry_name",
        "name": "protein_name",
        "OS": "organism_name",
        "OX": "organism_identifier",
        "GN": "gene_name",
        "PE": "protein_existence",
        "SV": "sequence_version",
    }

    @classmethod
    def parse(cls, header: str) -> ParsedHeader:
        """Parse a FASTA header string into a ParsedHeader object."""
        match = cls.header_pattern.match(header)
        if match is None:
            raise ValueError(f"Header does not match the UniProt pattern: {header}")
        fields = match.groupdict()

        for key in ["OS", "OX", "GN", "PE", "SV"]:
            if fields[key] is None:
                del fields[key]
        fields = {cls.field_names[key]: value for key, value in fields.items()}

        return ParsedHeader(fields["identifier"], header, fields)


class UniprotWriter:
    """Uniprot FASTA header writer."""

    field_names = {
        "db": "db",
        "id": "identifier",
        "entry": "entry_name",
        "name": "protein_name",
        "OS": "organism_name",
        "OX": "organism_identifier",
        "GN": "gene_name",
        "PE": "protein_existence",
        "SV": "sequence_version",
    }

    @classmethod
    def write(cls, parsed_header: AbstractParsedHeader) -> str:
        """Write a FASTA header string from a ParsedHeader object."""
        fields = parsed_header.header_fields
        header_entries = [
            f"{fields['db']}|{fields['identifier']}|{fields['entry_name']}",
            f"{fields['protein_name']}",
        ]
        for key in ["OS", "OX", "GN", "PE", "SV"]:
            field_name = cls.field_names[key]
            if field_name not in fields:
                continue
            header_entries.append(f"{key}={fields[field_name]}")
        return " ".join(header_entries)


class UniprotLikeParser:
    """A tolerant FASTA header parser for UniProt like headers."""

    field_pattern = re.compile(
        r"(?:(\s+OS=(?P<OS>[^=]+))|"
        r"(\s+OX=(?P<OX>\d+))|"
        r"(\s+GN=(?P<GN>\S+))|"
        r"(\s+PE=(?P<PE>\d))|"
        r"(\s+SV=(?P<SV>\d+)))*\s*$"
    )

    tag_names = {
        "OS": "organism_name",
        "OX": "organism_identifier",
        "GN": "gene_name",
        "PE": "protein_existence",
        "SV": "sequence_version",
    }

    @classmethod
    def parse(cls, header: str) -> ParsedHeader:
        """Parse a FASTA header string into a ParsedHeader object."""
        split_header = header.split(maxsplit=1)
        try:
            db, _id, entry = split_header[0].split("|")
        except ValueError:
            raise ValueError(
                f"Header does not match the minimal UniProt like pattern: {header}"
            )
        fields = {"db": db, "identifier": _id, "entry_name": entry}

        if len(split_header) == 1:
            return ParsedHeader(fields["identifier"], header, fields)

        description = split_header[1]
        tag_positions = [description.find(f"{tag}=") for tag in cls.tag_names]
        matched_start = sorted([num for num in tag_positions if num >= 0])
        matched_end = matched_start[1:] + [len(description)]

        if not matched_start:  # Description contains only the protein name
            fields["protein_name"] = description
        elif matched_start[0] != 0:  # Description contains protein name and tag fields
            fields["protein_name"] = description[: matched_start[0]].rstrip()

        if matched_start:  # Header contains tag fields
            for start, end in zip(matched_start, matched_end):
                matched_field = description[start:end].rstrip().split("=", maxsplit=1)
                fields[matched_field[0]] = matched_field[1]

        for old_tag, new_tag in cls.tag_names.items():
            if old_tag in fields:
                fields[new_tag] = fields.pop(old_tag)

        return ParsedHeader(fields["identifier"], header, fields)


class UniprotLikeWriter:
    """A tolerant FASTA header writer for UniProt like headers.

    In contrast to a strict UniProt header, the only required fields are the database,
    the identifier, and the entry name. The other fields are optional and can be
    omitted.
    """

    tag_names = {
        "OS": "organism_name",
        "OX": "organism_identifier",
        "GN": "gene_name",
        "PE": "protein_existence",
        "SV": "sequence_version",
    }

    @classmethod
    def write(cls, parsed_header: AbstractParsedHeader) -> str:
        """Write a FASTA header string from a ParsedHeader object."""
        fields = parsed_header.header_fields
        header_entries = [
            f"{fields['db']}|{fields['identifier']}|{fields['entry_name']}",
        ]
        if "protein_name" in fields:
            header_entries.append(f"{fields['protein_name']}")

        for key in ["OS", "OX", "GN", "PE", "SV"]:
            field_name = cls.tag_names[key]
            if field_name not in fields:
                continue
            header_entries.append(f"{key}={fields[field_name]}")
        return " ".join(header_entries)


def register_parser(name: str, parser: HeaderParser):
    """Register a custom parser by name."""
    PARSER_REGISTRY[name] = parser


def get_parser(parser_name: str) -> HeaderParser:
    """Get a registered parser by name."""
    return PARSER_REGISTRY[parser_name]


def register_writer(name: str, parser: HeaderWriter):
    """Register a custom writer by name."""
    WRITER_REGISTRY[name] = parser


def get_writer(parser_name: str) -> HeaderWriter:
    """Get a registered writer by name."""
    return WRITER_REGISTRY[parser_name]


PARSER_REGISTRY: dict[str, HeaderParser] = {
    "default": DefaultParser,
    "uniprot": UniprotParser,
    "uniprot_like": UniprotLikeParser,
}


WRITER_REGISTRY: dict[str, HeaderWriter] = {
    "default": DefaultWriter,
    "decoy": DecoyWriter,
    "uniprot": UniprotWriter,
    "uniprot_like": UniprotLikeWriter,
}



from copy import deepcopy
from profasta.db import ProteinDatabase


def create_decoy_db(
    db: ProteinDatabase, keep_nterm: bool = False, keep_nterm_methionine: bool = True
) -> ProteinDatabase:
    """Create a decoy database by reversing the sequences of the input database records.

    Args:
        db: The input database.
        keep_nterm: If True, keep the N-terminal residue in the same position. Default
            is False.
        keep_nterm_methionine: If True, keep the N-terminal residue in the same position
            if it is a methionine. Default is True.

    Returns:
        The decoy database.
    """
    decoy_db = ProteinDatabase()
    for protein in db:
        decoy_entry = deepcopy(db[protein])
        decoy_entry.sequence = reverse_sequence(
            decoy_entry.sequence,
            keep_nterm=keep_nterm,
            keep_nterm_methionine=keep_nterm_methionine,
        )
        decoy_db.add_entry(decoy_entry)
    return decoy_db


def reverse_sequence(
    sequence: str, keep_nterm: bool = False, keep_nterm_methionine: bool = True
) -> str:
    """Create a decoy sequence by reversing the input sequence.

    Args:
        sequence: The input sequence.
        keep_nterm: If True, keep the N-terminal residue in the same position. Default
            is False.
        keep_nterm_methionine: If True, keep the N-terminal residue in the same position
            if it is a methionine. Default is True.
    """
    if not sequence:
        return sequence
    if keep_nterm or (keep_nterm_methionine and sequence[0] == "M"):
        return sequence[0] + sequence[1:][::-1]
    return sequence[::-1]
